/****************************************************************************
**
** Definition of GenericTetris, a generic class for implementing Tetris.
**
** Author   : Eirik Eng
** Created  : 940126
** Modified : 991105 by Nicolas Hadacek
**
** Copyright (C) 1994 by Eirik Eng. All rights reserved.
**
**---------------------------------------------------------------------------
**
** The GenericTetris class is an abstract class that can be used to implement
** the well known game of Tetris. It is totally independent of any hardware
** platform or user interface mechanism. It has no notion of time, so you
** have to supply the timer ticks (or heartbeat) of the game. Apart from
** that this is a complete Tetris implementation.
**
** In the following it is assumed that the reader is familiar with the game
** of Tetris.
**
** The class operates on a grid of squares (referred to as the "board" below),
** where each of the different types of pieces in Tetris covers 4 squares. The
** width and height of the board can be specified in the constructor (default
** is 10x22). The coordinate (0,0) is at the TOP LEFT corner. The class
** assumes that it has total control over the board and uses this to optimize
** drawing of the board. If you need to update parts of the board
** (e.g. if you are using a window system), use the updateBoard() function.
**
** An implementation of the game must subclass from TetrisBoard and must
** implement these abstract functions:
**
**     virtual void drawSquare(int x,int y,int value)
**
**        This function is called when a square needs to be drawn
**        on the Tetris board. A value of 0 means that the square
**        should be erased. Values of 1 to 7 indicate the different
**        types of pieces. (Thus a minimal implementation can
**        draw 0 in one way and 1-7 in a second way). The x and y
**        values are coordinates on the Tetris board (see above).
**
**    virtual void gameOver()
**
**        This function is called when it is impossible to put a new
**        piece at the top of the board.
**
** To get a working minimal implementation of Tetris the following functions
** must be called from the user interface:
**
**     void startGame()
**
**         Clears the board and starts a new game.
**
**    void moveLeft()
**    void moveRight()
**    void rotateLeft()
**    void rotateRight()
**
**        The standard Tetris controls for moving and rotating the
**        falling pieces.
**
**    void dropDown();
**
**        Another Tetris control, drops the falling piece and calls the
**        virtual function pieceDropped(), whose default implementation is
**        to create a new block which appears at the top of the board.
**
**    void oneLineDown();
**
**        This is where you supply the timer ticks, or heartbeat, of the
**        game. This function moves the current falling piece one line down
**        on the board, or, if that cannot be done, calls the virtual
**        function pieceDropped() (see dropDown() above). The time between
**        each call to this function directly affects the difficulty of the
**        game. If you want to pause the game, simply block calls to the
**        user control functions above and stop calling this function (you
**        might want to call hideBoard() also).
**
** And that's it! There are several other public functions you can call
** and virtual functions you can overload to modify or extend the game.
**
** Do whatever you want with this code (i.e. the files gtetris.h, 
** gtetris.cpp, tpiece.h and tpiece.cpp). It is basically a weekend hack
** and it would bring joy to my heart if anyone in any way would find
** it useful.
**
** Nostalgia, comments and/or praise can be sent to: Eirik.Eng@troll.no
**
****************************************************************************/

#ifndef GTETRIS_H
#define GTETRIS_H

#include "matrix.h"

class GPiece;

class GenericTetris
{
 public:
    GenericTetris(uint width, uint height, bool graphical);
	virtual ~GenericTetris();

	virtual void init(int seed);
	virtual void start();
	
	void moveLeft(int steps = 1)  { moveTo(-steps, 0); }
    void moveRight(int steps = 1) { moveTo(steps, 0); }
    void rotateLeft()             { rotate(TRUE); }
    void rotateRight()            { rotate(FALSE); }
	virtual void oneLineDown();
	virtual void dropDown();

	uint  score() const        { return _score; }
    uint  level() const        { return _level; }
	uint  nbClearLines() const { return _nbClearLines; }
	uint  nbRemoved() const    { return _nbRemoved; }
	bool graphical() const     { return _graphical; }

	uint firstClearLine() const { return _matrix.height() - _nbClearLines; }
	uint currentPos() const     { return _currentPos; }
	int  currentLine() const    { return _currentLine; }

	virtual GPiece *createPiece() const = 0;
	const GPiece *nextPiece() const    { return _nextPiece; }
	const GPiece *currentPiece() const { return _currentPiece; }

	const Matrix<uint> &matrix() const { return _matrix; }

 protected:
	Matrix<uint> matrix() { return _matrix; }

	int moveTo(int decX, int decY);
	virtual void pieceDropped(uint dropHeight) = 0;
	virtual void gameOver() = 0;
	virtual void newPiece();
	bool canPosition(uint newPos, uint newLine, const GPiece *newPiece) const;
	void partialCopy(const GenericTetris &gt);
	virtual void gluePiece() = 0;
	void computeClearLines();

	virtual void updateRemoved(uint newNbRemoved) { _nbRemoved = newNbRemoved;}
    virtual void updateScore(uint newScore)       { _score = newScore; }
    virtual void updateLevel(uint newLevel)       { _level = newLevel; }
	
	void setCurrentPos(uint pos)   { _currentPos = pos; }
	void setCurrentLine(uint line) { _currentLine = line; }
	void addToClearLines(int nb)   { _nbClearLines += nb; }
	
	// graphic methods
	int X(int i) const { return i; }
	int Y(int j) const { return _matrix.height() - 1 - j; }

	void hideBoard() { showBoard(TRUE); }
    void showBoard(bool hide = FALSE);
	void updateBoard()
		{ updateBoard(0, 0, _matrix.width()-1, _matrix.height()-1); }
	void updateBoard(uint x1, uint y1, uint x2, uint y2);
	virtual void showNextPiece(bool) {}
	void hideNextPiece() { showNextPiece(TRUE); }
	virtual void showPieceConfig(bool) {}
	void hidePieceConfig() { showPieceConfig(TRUE); }
	virtual void drawBlock(uint i, uint j, int color) = 0;
	void showPiece()
		{ showPiece(0, 0, _matrix.width()-1, _matrix.height()-1); }
	void showPiece(uint x1, uint y1, uint x2, uint y2);	
	void setAndDrawBlock(uint i, uint j, uint value);
	bool hidden() const { return _hidden; }
	
 private:
	void clear();
	void rotate(bool left);
    void optimizedMove(uint newPos, uint newLine, const GPiece *newPiece);

	uint         _score, _level, _nbRemoved, _nbClearLines, _currentPos;
	int          _currentLine;
	GPiece      *_nextPiece, *_currentPiece;
	bool         _graphical, _hidden;
	Matrix<uint> _matrix;
};

#endif
