#include "multi.h"

#include <kapp.h>
#include <klocale.h>

#include "askgame.h"
#include "askname.h"
#include "defines.h"
#include "types.h"
#include "meeting.h"
#include "setlocal.h"
#include "multi_private.h"
#include <qmessagebox.h>

/*****************************************************************************/
/* Constructor & Destructor                                                  */
/*****************************************************************************/
MultiPlayerInterface::MultiPlayerInterface(const GameInfo &_gameInfo, KAccel *_kacc,
										   QWidget *parent, const char *name)
: QWidget(parent, name), internal(0), gameInfo(_gameInfo),
  nbLocalHumans(0), hbl(this), kacc(_kacc)
{
	ASSERT( _gameInfo.maxNbLocalPlayers>=1 );
	
	keys.setAutoDelete(TRUE);
	hbl.setResizeMode( QLayout::Fixed );
}

MultiPlayerInterface::~MultiPlayerInterface()
{
	delete internal;
}

/*****************************************************************************/
/* Game creation                                                             */
/*****************************************************************************/
void MultiPlayerInterface::dialog()
{
	if (internal) stop();
	
	// ask CREATE or JOIN
	AskGame ag(gameInfo.trGameName, this);
	bool server;
	switch ( ag.exec() ) {
	 case AskGame::Cancel: return;
	 case AskGame::Create: server = TRUE; break;
	 case AskGame::Join:   server = FALSE; break;
	}
	
	// ask game details
	ConnectionData cd;
	AskName an(gameInfo, server, cd, this);
	connect(&an, SIGNAL(addKeys(uint)), SLOT(addKeys(uint)));
	if ( !an.exec() ) return;

	// net meeting
	QArray<RemoteHostData> rhd;
	if (cd.netPlayersAllowed) {
		cId id(gameInfo.gameName, gameInfo.gameId);
		OptionWidget *ow = newOptionWidget(server);
		NetMeeting *nm;
		if (server) nm = new ServerNetMeeting(gameInfo.trGameName, id, cd.rhd, ow, rhd, this);
		else nm = new ClientNetMeeting(gameInfo.trGameName, id, cd.rhd, ow, this);
		int res = nm->exec();
		if ( ow && res ) ow->saveData();
		delete ow;
		delete nm;
		if (!res) return;
	}
	
	createLocalGame(cd.rhd.gd, server, rhd.size());
	if (server) createServerGame(rhd);
	else createClientGame(cd.rhd);
}

void MultiPlayerInterface::specialLocalGame(uint nbHumans, uint nbComputers)
{
	if (internal) stop();
	
	GameData gd;
	gd.nbHumans = nbHumans;
	gd.nbComputers = nbComputers;
	gd.name = SetLocal::playerName(1, FALSE);
	ASSERT( gameInfo.maxNbLocalPlayers>=gd.nbPlayers() );
	ASSERT( gameInfo.AIAllowed || nbComputers==0 );
	createLocalGame(gd, TRUE, FALSE);
	
	QArray<RemoteHostData> rhd;
	createServerGame(rhd);
}

void MultiPlayerInterface::createServerGame(const QArray<RemoteHostData> &rhd)
{
	delete internal;
	internal = (rhd.size()
				? (Local *)new NetworkServer(this, boards, rhd, gameInfo.interval)
				: (Local *)new LocalServer(this, boards, gameInfo.interval));
	init(TRUE);
}

void MultiPlayerInterface::createClientGame(const RemoteHostData &rhd)
{
	delete internal;
	QArray<RemoteHostData> a(1);
	a[0] = rhd;
	internal = new Client(this, boards, a);
	init(FALSE);
}

void MultiPlayerInterface::createLocalGame(const GameData &gd, bool _server,
										   bool network)
{
	server = _server;
	
	// basic assumption here : the human players are first and 
	// then are computer players (could be change some time in the future but
	// carefully)

	// remove unused keys 
	for (uint i=gd.nbHumans; i<nbLocalHumans; i++) removeKeys(i);

	// add or remove boards
	uint old_gs = boards.size();
	ASSERT( gd.nbPlayers()>=1 );
	for (uint i=gd.nbPlayers(); i<old_gs; i++) delete boards[i];

	boards.resize(gd.nbPlayers());
	for(uint i=old_gs; i<boards.size(); i++) {
		boards[i] = newBoard(i);
		hbl.addWidget(boards[i]);
		boards[i]->show();
	}

	// add and reconnect used keys
	for (uint i=nbLocalHumans; i<gd.nbHumans; i++) addAndConnectKeys(i);
	nbLocalHumans = gd.nbHumans;
	debug("nb local humanes %i", nbLocalHumans);

	// init local boards
	for(uint i=0; i<boards.size(); i++) {
		disconnect(boards[i], SIGNAL(activateKeys(bool)),
				   this, SLOT(activateKeys(bool)));
		if ( i<nbLocalHumans ) {
			debug("connect %i", i);
			connect(boards[i], SIGNAL(activateKeys(bool)),
					SLOT(activateKeys(bool)));
		}
		boards[i]->init(i>=nbLocalHumans, network || gd.nbPlayers()>1,
						server);
	}
}

/*****************************************************************************/
/* Key management                                                            */
/*****************************************************************************/
void MultiPlayerInterface::addKey(const QString &descr, const QString &action,
								  const QString &member,
								  const QArray<uint> &predefinedKeyCodes)
{
	// sanity check
	for(KeyData *k=keys.first(); k!=0; k = keys.next())
		ASSERT( k->action!=action );

	KeyData *k = new KeyData;
	keys.append(k);
	k->descr   = descr;
	k->action  = action;
	k->member  = member;
	k->codes   = predefinedKeyCodes.copy();
}

void MultiPlayerInterface::addKey(const QString &descr, const QString &action,
								  const QString &member, uint nbPredefined,
								  const char **predefinedKeyCodes)
{
	QArray<uint> pkc(nbPredefined);
	for (uint k=0; k<nbPredefined; k++)
		pkc[k] = KAccel::stringToKey(predefinedKeyCodes[k]);
	addKey(descr, action, member, pkc);
}

void MultiPlayerInterface::addAndConnectKeys(uint human_i, bool remove)
{
	QString s, d, a;
	s = createKeyPrefix(human_i);
	for (KeyData *k = keys.first(); k!=0; k = keys.next()) {
		d = s + k->descr;
		a = s + k->action;
		if (remove) kacc->removeItem(a); // auto disconnect (?)
		else {
			kacc->insertItem(d, a, (human_i<k->codes.size() ? 
									k->codes[human_i] : findUnusedKey()));
			kacc->connectItem(a, boards[human_i], k->member, FALSE);
		}
	}
}

uint MultiPlayerInterface::findUnusedKey()
{
	return KAccel::stringToKey("A"); //###### FIXME
}

void MultiPlayerInterface::activateKeys(bool activate)
{
	uint i;
	for (i=0; i<nbLocalHumans; i++)
		if ( sender()==boards[i] ) break;
	ASSERT( i<nbLocalHumans );
	QString s = createKeyPrefix(i);
	for(KeyData *k = keys.first(); k!=0; k = keys.next())
		kacc->setItemEnabled(s + k->action, activate);
}

QString MultiPlayerInterface::createKeyPrefix(uint i)
{
	QString s;
	s.setNum(i);
	return s + " ";
}

/*****************************************************************************/
/* Misc. functions                                                           */
/*****************************************************************************/
uint MultiPlayerInterface::nbPlayers() const
{
	ASSERT(server);
	return internal->nbPlayers();
}

QDataStream &MultiPlayerInterface::readingStream(uint i) const
{
	ASSERT(server);
	return internal->ioBuffer(i)->reading;
}

QDataStream &MultiPlayerInterface::writingStream(uint i) const
{
	ASSERT(server);
	return internal->ioBuffer(i)->writing;
}

QDataStream &MultiPlayerInterface::dataToClientsStream() const
{
	ASSERT(server);
	return *internal->globalStream();
}

void MultiPlayerInterface::immediateWrite()
{
	ASSERT(server);
	internal->writeData();
}

void MultiPlayerInterface::error(const QString &msg)
{
	delete internal;
	internal = 0;
	debug(QString("ERROR : %1").arg(msg));
	ERROR(msg);
}
