#ifndef MULTI_H
#define MULTI_H

#include <qwidget.h>
#include <qlayout.h>
#include <qtimer.h>
#include <kapp.h>
#include <kaccel.h>

/**
 * The LocalBoard class is the base widget from which each individual
 * board should inheritate ; you must implement its virtual methods.
 */
class LocalBoard : public QWidget
{
 Q_OBJECT
		
 public:
    LocalBoard(QWidget *parent, const char *name=0)
	: QWidget(parent, name) {}
	virtual ~LocalBoard() {}

	/**
	 * This method is called once at the board creation.
	 * @param AI is TRUE if the player is not human.
	 * @param multiplayers is TRUE if the game is not a single player game.
	 */
	virtual void init(bool AI, bool multiplayers, bool server) = 0;
	
	/** 
	 * Put data on the stream.
	 * 
	 * This method is the communication way out. The data given here will
	 * be the only information that will go to the server.
	 */
	virtual void dataOut(QDataStream &) = 0;
	
	/** 
	 * Get data from the stream.
	 * 
	 * This method is the communication way in. The data given here will be 
	 * the only information that you will receive from the server.
	 */
	virtual void dataIn(QDataStream &) = 0;

 signals:
	/**
	 * Call this signal to enable/disable the keys associated with a board.
	 */
	void activateKeys(bool);
};

/** 
 * The OptionWidget is a base widget for the option widget in the
 * "netmeeting" dialog. This option widget is optional (!). 
 *
 * For example you will have :	 
 * <PRE>
 *  class MyOptionWidget : public OptionWidget { ... };
 *  class MyMultiPlayerInterface : public MultiPlayerInterface { ... };
 *   
 *  OptionWidget *MyMultiPlayerInterface::newOptionWidget(bool server) const
 *  { return new MyOptionWidget(server); };
 * </PRE>
 *
 * The option widget must have two different behaviours for server and
 * clients. The server is able to change the options but the clients are only
 * able to see them. The library will catch the @ref #changed signal from
 * the option widget and will send the changes to the clients. It uses the
 * @ref #dataOut to obtain the data from the
 * server option dialog and then on the client side, it sets the new data with
 * the method @ref #dataIn.
 * You must implement this three methods to have useful option widgets and be
 * careful to emit the signal @ref #changed whenever the server widget is
 * changed. In addition you'll need to implement the method @ref #saveData
 * to save the configuration in the config file ; so that it will be available
 * for the initialisation of @ref LocalBoard.
 * It seems a good idea that the widget have the same layout
 * on both (server and client) sides but with the inner widgets all disabled 
 * on the client side.
 */
class OptionWidget : public QWidget
{
 Q_OBJECT
		
 public:
	OptionWidget(bool Server, QWidget *parent = 0, const char *name = 0)
	: QWidget(parent, name), server(Server) {}
	virtual ~OptionWidget() {}

	bool isServer() const { return server; }
	
	/** 
	 * This method is used on the client side to set the data coming from
	 * the server widget.
	 */
	virtual void dataIn(QDataStream &s) = 0;
	
	/** This method is used on the server side to get the data. */
	virtual void dataOut(QDataStream &s) const = 0;

	/** 
	 * When the game will begin (ie when the "netmeeting" is over and the
	 * server has press the "start game" button) this method will be called
	 * (for clients and server). It must save the settings in the config
	 * file.
	 */
	virtual void saveData() = 0;
	
 signals:
	/** 
	 * This signal must be called each time options are changed 
	 * (by the server).
	 */
	void changed();
	
 private:
	bool server;
};

class Local;
class GameData;
class RemoteHostData;

/**
 * This structure contains information about the game
 * configuration.
 */
typedef struct {
	/** Internationalized name of the game (e.g. i18n("ksirtet")). */
	QString trGameName;
	
	/** The game name used for identification (e.g. ksirtet).
	 *  So that a network client for a car race game cannot join
	 *  in a "sirtet" game.
	 */
	QString	gameName;
	
	/** The game version id used for identification (e.g. "4").
	 *  You should change this id when the game is made incompatible
	 *  with previous version. (changes in data for example).
	 */
    QString gameId;
	
	/** Maximum number of local players. */
	uint maxNbLocalPlayers;
	
	/** Interval (in msec.) between data exchange. */
	uint interval;
	
	/** If there are built-in artificial intelligences that can play. */
	bool AIAllowed;
	
	/** Slot for player/AI additional configuration. These must be SLOTs which
	 * take an unsigned int as parameter. It must open a setting
	 * dialog for the corresponding local player/computer and save the
	 * new settings in the config file. It should probably create a group
	 * with the given number in its name.
	 * If such a pointer is set to 0 : it means there is no perticular
	 * setting.
	 */
	 const char *playerSetSlot, *AISetSlot;
} GameInfo;

/** 
 * The MultiPlayerInterface class is useful for multiplayer game
 * management. Each game is represented by a class you have inherited
 * from the @ref LocalBoard class.
 * 
 * Multiplayers games can take place with several (humans or eventually
 * AIs) players on the same computer (they use the same keyboard and have
 * each a @ref LocalBoard widget on the screen) or/and network players.
 *
 * This class is intended to do all the hard work of sending/receiving data
 * between the players and to send the keyboard events to the right
 * @ref LocalBoard. So multiplayers game should be completely transparent
 * from your point of view.
 * 
 * Note : The data exchange is done in background with a timer calling at given
 * intervals the read/write methods. Obviously this kind of things can be done
 * easily with threads but I have no experience with this kind of programming
 * and not all people have thread libraries and a thread-safe system.
 */

class MultiPlayerInterface : public QWidget
{
 Q_OBJECT
		
 public:
	/** Constructor which takes a GameInfo struct as parameter. */
    MultiPlayerInterface(const GameInfo &gameInfo, KAccel *kacc,
						 QWidget *parent = 0, const char *name = 0);
	virtual ~MultiPlayerInterface();
	
 public slots:
	/** Create a single player game for a human being.
	  * Call @ref stop if a game is already created. */
	void singleHuman()     { specialLocalGame(1, 0); }
	/** Create a local game opposing two human beings.
	  * Call @ref stop if a game is already created. */
	void humanVsHuman()    { specialLocalGame(2, 0); }
	/** Create a local game opposing a human with an AI.
	  * Call @ref stop if a game is already created. */
	void humanVsComputer() { specialLocalGame(1, 1); }

	/** Open a dialog to create a multiplayer game.
	  * Call @ref stop if a game is already created. */
	void dialog();
	
 public:
	/** Called when a new game is created. At this point
	 *  the number of players is known. */
	virtual void init(bool /*server*/) {}
	
	/** Called just before a new game is created (called by
	  * singleHuman, humanVsHuman, humanVsComputer and dialog). */
	virtual void stop() {}
	
	/** 
	 * Add a function that is handled by the keyboard.
	 *
	 * @param functionName is the string that will appear in the keys 
	 * configuration dialogs : it will be preceded by the player range.
	 *
	 * @param member is the slot name which the key event must call.
	 *
	 * @param nbPredefined is the number of predefined keycodes.
	 *
	 * @param predefinedKeyCodes contains the predefines
	 * keycodes. They are used as default if no entry are present in the
	 * configuration files. 
	 * 	 
	 * The function/key connections are disabled ; you must call the 
	 * @ref LocalBoard::activateKeys method to enable them.
	 * 	 
	 * An example follows :
	 * <PRE>
	 *  uint keycodes[2];
	 *  keycodes[0] = stringToKey("A");
	 *  keycodes[1] = stringToKey("Space");
	 *  addKey(i18n("Fire"), "Fire", SLOT(fire()), 2, keycodes);
	 * <PRE>
	 * In this example when the first player hit the A key, it will "fire" in
	 * his board. When the second player hit the bar space, it will do
	 * the same in his board. Maybe the first player missile will
	 * touch the second player ship ...
	 * Also note that default keys for the first player are used in single
	 * player game. [this can change in the future since keyboard use
     * can be different with multiple players]
	 */
	void addKey(const QString &descr, const QString &action,
				const QString &member, const QArray<uint> &predefinedkeyCodes);

	/** For convenience. */
	void addKey(const QString &descr, const QString &action,
				const QString &member,
				uint nbPredefined, const char **predefinedKeyCodes);
	
	/** @return the total number of players
	  * Do not call from client !. 
	  */
	uint nbPlayers() const;

	/** 
	 * Create a new @ref LocalGameBoard.
	 * 
	 * @param i is the game index that goes from 0 to the number of
	 * local players : it can be used to retrieve configuration settings.
	 */
	virtual LocalBoard *newBoard(uint i) = 0;

	/** 
	 * This method must read data from each client with method
	 * @ref readingStream, do the needed treatement
	 * (for instance which players has lost, which data to be resent, ...) and
	 * then write the useful data to each client with method
	 * @ref writingStream.
	 *
	 * NB: this method is also called for single player games but
	 * you probably only want to check for game over condition (it depends
	 * on game implementation that you really return data to the board).
	 * Doing a write in single player game is not mandatory.
	 */
	virtual void treatData() = 0;
	
	/** @return the reading stream for board #i.
	  * Do not call from client ! 
	  */
	QDataStream &readingStream(uint i) const;
	/** @return the writing stream for board #i.
	  * Do not call from client !
	  */
	QDataStream &writingStream(uint i) const;

	/**
	 * Read data sent from server to clients "MultiplayersInterface" 
	 * (this data is not adressed to boards).
	 . These are meta data that are not directly used in game.
	 * It can be used to display "game over" infos for all
	 * local games.
	 * NB: the use of this methode is optional.
	 */
	virtual void dataFromServer(QDataStream &) {}
	
	/** Used by the server to write meta data to clients.
	 *  NB: the use of this method is optional.
	 *  Do not call from client !
	 */
	QDataStream &dataToClientsStream() const;

	/** Write immediately data to clients and boards.
	  * It is unlike the normal exchange which is driven
	  * by the timer of the server. Beware of possible
	  * interactions.
	  * 
	  * Do not call from client !
	  */
	void immediateWrite();
	
	/** Display an error and destroy current game. */
	void error(const QString &s);
	
	/** 
	 * This method should be overload if an option widget is used in the
	 * multiplayer setting dialog (in the "netmeeting" dialog). By default a
	 * null pointer is returned and so no option widget is shown.
	 * The option widget must be inherited from the @ref OptionWidget class.
	 */
	virtual OptionWidget *newOptionWidget(bool /*server*/) const { return 0; };
	
 private slots:
	void activateKeys(bool);
	void addKeys(uint i) { addAndConnectKeys(i); } // #### hacky : see Askname

 private:
		typedef struct {
		QString      action, descr, member;
		QArray<uint> codes;
	} KeyData;

	Local               *internal;
	GameInfo             gameInfo;
	QArray<LocalBoard *> boards;
	uint                 nbLocalHumans;
	QHBoxLayout          hbl;
	QList<KeyData>       keys;
	KAccel              *kacc;
	bool                 server;

	void createServerGame(const QArray<RemoteHostData> &rhd);
	void createClientGame(const RemoteHostData &rhd);
	void createLocalGame(const GameData &gd, bool server, bool network);
	
	void specialLocalGame(uint nbHumans, uint nbComputers);
	
	void addAndConnectKeys(uint i, bool remove = FALSE);
	void removeKeys(uint i) { addAndConnectKeys(i, TRUE); }
	uint findUnusedKey();
	static QString createKeyPrefix(uint i);
};

#endif // MULTI_H
