#include "pline.h"

#include <qfont.h>
#include <qpushbutton.h>
#include <klocale.h>
#include "defines.h"

#define THIN_BORDER 4

MeetingLine::MeetingLine(bool serverIsReader, bool isOwner,
					   QWidget *parent, const char *name)
: QFrame(parent, name)
{
	setFrameStyle(Panel | Raised);
	
	// Top layout
	hbl = new QHBoxLayout(this, THIN_BORDER + frameWidth());
	hbl->setResizeMode(QLayout::Fixed);

	/* TriCheckBox */
	tcb = new MeetingCheckBox(MeetingCheckBox::Ready, serverIsReader, this);
	if ( !XOR(isOwner, serverIsReader) ) tcb->setEnabled(FALSE);
	else connect(tcb, SIGNAL(changed(int)), SLOT(_typeChanged(int)));
	hbl->addWidget(tcb);

    /* Name */
	lname = new QLabel(" ", this);
	lname->setAlignment(AlignCenter);
	lname->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	lname->setLineWidth(2);
	lname->setMidLineWidth(3);
	QFont f = lname->font();
	f.setBold(TRUE);
	lname->setFont(f);
	lname->setFixedSize(lname->fontMetrics().maxWidth()*NAME_MAX_LENGTH,
						lname->sizeHint().height());
	hbl->addWidget(lname);
	hbl->addStretch(1);

	/* Nb humans */
	labH = new QLabel(this);
	hbl->addWidget(labH);
	
	/* Nb AIs */
	labAI = new QLabel(this);
	hbl->addWidget(labAI);
}

void MeetingLine::setData(const ExtData &ed)
{
	bds = ed.bds;
	uint nbh = 0, nba = 0;
	for (uint i=0; i<bds.count(); i++) {
		if ( bds[i].type==PlayerCheckBox::Human ) nbh++;
		else if ( bds[i].type==PlayerCheckBox::AI ) nba++;
	}
	labH->setText(i18n("Hu=%1").arg(nbh));
	labAI->setText(i18n("AI=%1").arg(nba));
 	lname->setText(bds[0].name);
	setType(ed.type);
	setText(ed.text);
}

void MeetingLine::data(ExtData &ed) const
{
	ed.bds  = bds;
	ed.type = tcb->type();
	ed.text = text();
}

/*****************************************************************************/
EditMeetingLine::EditMeetingLine(bool readerIsServer,
							   QWidget *parent, const char *name)
: MeetingLine(readerIsServer, TRUE, parent, name)
{
	// Talker
	qle = new QLineEdit(" ", this);
	qle->setMaxLength(TALKER_MAX_LENGTH);
	qle->setFont( QFont("fixed", 12, QFont::Bold) );
	qle->setFixedSize(qle->fontMetrics().maxWidth()*TALKER_MAX_LENGTH,
					  qle->sizeHint().height()); // to get the same sizeHint
		                                          // as a label
	qle->setText("");
	connect(qle, SIGNAL(textChanged(const QString &)), 
			SLOT(_textChanged(const QString &)));
	hbl->addWidget(qle);
}

/*****************************************************************************/
ShowMeetingLine::ShowMeetingLine(bool readerIsServer,
							   QWidget *parent, const char *name)
: MeetingLine(readerIsServer, FALSE, parent, name)
{
	// Talker
	lab = new QLabel(" ", this);
	lab->setFrameStyle(QFrame::Panel | QFrame::Sunken);
	lab->setLineWidth(2);
	lab->setMidLineWidth(3);
	lab->setFont( QFont("fixed", 12, QFont::Bold) );
	lab->setFixedSize( lab->fontMetrics().maxWidth()*(TALKER_MAX_LENGTH+2),
					   lab->sizeHint().height() );
	hbl->addWidget(lab);
}

/*****************************************************************************/
PlayerLine::PlayerLine(PlayerCheckBox::Type type, const QString &txt,
					   bool humanSetting, bool AISetting,
					   bool canBeEmpty, bool acceptAI,
					   QWidget *parent, const char *name)
: QFrame(parent, name), hs(humanSetting), as(AISetting)
{
	setFrameStyle(Panel | Raised);
	
	// Top layout
	QHBoxLayout *hbl;
	hbl = new QHBoxLayout(this, THIN_BORDER + frameWidth());
	hbl->setResizeMode(QLayout::Fixed);

	/* CheckBox */
	pcb = new PlayerCheckBox(type, canBeEmpty, acceptAI, this);
	connect(pcb, SIGNAL(changed(int)), SLOT(typeChangedSlot(int)));
	hbl->addWidget(pcb);

    /* Name */
	edit = new QLineEdit(txt, this);
	edit->setMaxLength(NAME_MAX_LENGTH);
	edit->setFixedSize(edit->fontMetrics().maxWidth()*(NAME_MAX_LENGTH+2),
					   edit->sizeHint().height());
	hbl->addWidget(edit);

	/* settings button */
	setting = new QPushButton(i18n("Settings"), this);
	connect(setting, SIGNAL(clicked()), SLOT(setSlot()));
	hbl->addWidget(setting);
	
	typeChangedSlot(type);
}

void PlayerLine::typeChangedSlot(int t)
{
	edit->setEnabled(type()!=PlayerCheckBox::Empty);
	setting->setEnabled( (type()==PlayerCheckBox::Human && hs)
						 || (type()==PlayerCheckBox::AI && as) );
	emit typeChanged(t);
}

void PlayerLine::setSlot()
{
	if ( type()==PlayerCheckBox::Human ) emit setHuman();
	else emit setAI();
}

/*****************************************************************************/
GWidgetList::GWidgetList(uint _interval, QWidget *parent, const char * name)
: QWidget(parent, name), interval(_interval)
{
	widgets.setAutoDelete(TRUE);
}

uint GWidgetList::maxWidth() const
{
	int wd = 0;
	QListIterator<QWidget> it(widgets);
	for (; it.current(); ++it)
		wd = QMAX(wd, it.current()->sizeHint().width());
	return (uint)wd;
}

uint GWidgetList::totalHeight() const
{
	uint h = 0, i = 0;
	QListIterator<QWidget> it(widgets);
	for (; it.current(); ++it) {
		h += (uint)it.current()->sizeHint().height() + (i ? interval : 0);
		i++;
	}
	return h;
}

QSize GWidgetList::sizeHint() const
{
	return QSize(maxWidth(), totalHeight());
}

QSizePolicy GWidgetList::sizePolicy() const
{
	return QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
}

void GWidgetList::append(QWidget *wi)
{
	wi->move(0, totalHeight() + (widgets.count() ? interval : 0));
	wi->show();
	widgets.append(wi);
	updateGeometry();
}

void GWidgetList::remove(uint i)
{
	int dec = widgets.at(i)->height() + interval;
	widgets.remove(i);
	QListIterator<QWidget> it(widgets);
	for(; it.current(); ++it) it.current()->move(0, it.current()->y()-dec);
	updateGeometry();
}
