#include "misc_ui.h"

#include <qtimer.h>
#include "gpiece.h"
#include "board.h"

/*****************************************************************************/
ShowNextPiece::ShowNextPiece(const Board *board, QWidget *parent,
							 const char *name)
: QFrame(parent, name), _shown(FALSE), _p(this), _board(board)
{
	setFrameStyle( QFrame::Panel | QFrame::Sunken );
}

QSize ShowNextPiece::sizeHint() const
{
	GPieceInfo *info = _board->createPieceInfo();
	return QSize((info->maxWidth()+2) * _board->blockWidth(),
				 (info->maxHeight()+2) * _board->blockHeight());
	delete info;
}

QSizePolicy ShowNextPiece::sizePolicy() const
{
	return QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
}

void ShowNextPiece::clearAndShow()
{
	if (_shown) repaint();
	else reveal(TRUE);
}	

void ShowNextPiece::reveal(bool reveal)
{
	if ( _shown==reveal ) return;
	_shown = reveal;
	draw(!_shown);
}

void ShowNextPiece::redraw(bool clear)
{
	if (_shown) draw(clear);
}

void ShowNextPiece::paintEvent(QPaintEvent *)
{
	drawFrame(&_p);
	if (_shown) draw(FALSE);
}

void ShowNextPiece::draw(bool erase)
{
	const GPiece *piece = _board->nextPiece();
	uint pw = piece->width()  * _board->blockWidth();
	uint ph = piece->height() * _board->blockHeight();
	uint xm = piece->minX()   * _board->blockWidth();
	uint ym = piece->minY()   * _board->blockHeight();
	int xOffset = (width()  - pw)/2 - xm;
	int yOffset = (height() - ph)/2 - ym;
	QRect r(xOffset, yOffset, _board->blockWidth(), _board->blockHeight());
		
	QRect tmp;
	for (uint k=0; k<piece->nbBlocks(); k++) {
		tmp = r;
		tmp.moveBy(piece->pos(k, 0)   * _board->blockWidth(),
				   -piece->line(k, 0) * _board->blockHeight());
		_board->drawBlock(&_p, tmp, (erase ? -1 : piece->value(k)));
	}
}

/*****************************************************************************/
#define SHADOW_HEIGHT 10

Shadow::Shadow(const Board *board, int xOffset,
			   QWidget *parent, const char *name)
: QWidget(parent, name), _shown(FALSE), _p(this),
  _xOffset(xOffset), _board(board)
{
	_state.resize(_board->matrix().width());
	_state.fill(FALSE);
}

QSize Shadow::sizeHint() const
{
	return QSize(_xOffset + _board->matrix().width() * _board->blockWidth(),
				 SHADOW_HEIGHT);
}

QSizePolicy Shadow::sizePolicy() const
{
	return QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
}

void Shadow::reveal(bool reveal)
{
	if ( _shown==reveal ) return;
	_shown = reveal;
	draw(!_shown, FALSE);
}

void Shadow::redraw(bool clear)
{
	if (clear) return; // the state is saved so no clearing needed
	if (_shown) draw(FALSE, TRUE);
}

void Shadow::paintEvent(QPaintEvent *)
{
	if (_shown) draw(FALSE, FALSE);
}

void Shadow::draw(bool erase, bool smart)
{
	const GPiece *piece = _board->currentPiece();
	uint pf = piece->minX() + _board->currentPos();
	uint pl = pf + piece->width() - 1;

	for (uint i=0; i<_board->matrix().width(); i++) {
		if ( (i<pf || i>pl) || erase ) {
			if (_state[i]) {   // need deletion
			_state[i] = FALSE;
			drawBlock(i, TRUE);
			}
		} else if ( !_state[i] || !smart ) { // need drawing
				_state[i] = TRUE;
				drawBlock(i, FALSE);
		}
	}
}

void Shadow::drawBlock(uint i, bool erase)
{
	QRect r(_xOffset + i * _board->blockWidth() + 1 , 0,
			_board->blockWidth() - 2, SHADOW_HEIGHT);
	QColor col = _board->color(&_p, (erase ? -1 : 0));
	_p.setBrush(col);
	_p.setPen(col);
	_p.drawRect(r);
}

/*****************************************************************************/
#define GI_WIDTH     15
#define GI_HEIGHT    11
#define ARROW_HEIGHT  3
#define ARROW_WIDTH   7

GiftShower::GiftShower(uint timeout, QWidget *parent, const char *name)
: QWidget(parent, name), _timeout(timeout), _state(FALSE)
{}

QSize GiftShower::sizeHint() const
{
	return QSize(GI_WIDTH, GI_HEIGHT);
}

QSizePolicy GiftShower::sizePolicy() const
{
	return QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
}

void GiftShower::set()
{
	if (_state) return;
	_state = TRUE;
	repaint();
	QTimer::singleShot(_timeout, this, SLOT(timeout()));
}

void GiftShower::timeout()
{
	_state = FALSE;
	repaint();
}

void GiftShower::paintEvent(QPaintEvent *)
{
	QPainter p(this);
	p.setBrush((_state ? yellow : black));
	p.setPen(black);
	p.drawRect(0, (height()-ARROW_HEIGHT)/2, ARROW_WIDTH, ARROW_HEIGHT);
	QPointArray a(3);
	a.setPoint(0, ARROW_WIDTH, 0);
	a.setPoint(1, width()-1, height()/2);
	a.setPoint(2, ARROW_WIDTH, height()-1);
	p.drawPolygon(a);
}
	
/*****************************************************************************/
#define LED_WIDTH  15
#define LED_HEIGHT 15

// #### FIXME after KLed fixes commit
class Led : public QWidget // KLed
{
 public:
	Led(const QColor &c, QWidget *parent, const char *name=0)
		: QWidget(parent, name), col(c), _on(FALSE) {}
//		: KLed(c, Off, flat, Rectangular, parent, name) {}

	QSizePolicy sizePolicy() const
		{ return QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed); }
	QSize sizeHint() const { return QSize(LED_WIDTH, LED_HEIGHT); }

	void on()  { if (!_on) { _on = TRUE; repaint(); } }
	void off() { if (_on)  {_on = FALSE; repaint(); } }
	void setColor(const QColor &c) { if (c!=col) { col = c; repaint(); } }

 protected:
	void paintEvent(QPaintEvent *) {
		QPainter p(this);
		p.setBrush((_on ? col.light() : col.dark()));
		p.setPen(black);
		p.drawEllipse(0, 0, width(), height());
	}

 private:
	QColor col;
	bool   _on;
};

#define LED_SPACING 5


GiftPool::GiftPool(uint t, uint nbLeds,
				   QWidget *parent, const char *name)
: QHBox(parent, name), _timeout(t), nb(0), ready(false)
{
	setSpacing(LED_SPACING);
	leds.resize(nbLeds);
	for (uint i=0; i<nbLeds; i++) leds[i] = new Led(yellow, this);
}

QSize GiftPool::sizeHint() const
{
	QSize s = (leds.size() ? leds[0]->sizeHint() : QSize());
	return QSize((s.width()+LED_SPACING)*leds.size()-LED_SPACING, s.height());
}

QSizePolicy GiftPool::sizePolicy() const
{
	return QSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
}

void GiftPool::put(uint n)
{
	if ( n==0 ) return;
	if ( nb==0 ) QTimer::singleShot(_timeout, this, SLOT(timeout()));
	uint e = QMIN(nb+n, leds.size());
	for (uint i=nb; i<e; i++) leds[i]->on();
	uint f = QMIN(nb+n-e, leds.size());
	for (uint i=0; i<f; i++) leds[i]->setColor(red);
	nb += n;
}

uint GiftPool::take()
{
	ASSERT(ready);
	ready = FALSE;
	uint n = nb;
	nb = 0;
	for (uint i=0; i<leds.size(); i++) {
		leds[i]->setColor(yellow);
		leds[i]->off();
	}
	return n;
}
