#include "tl_board.h"

#include "defines.h"

#define TL_WIDTH  10
#define TL_HEIGHT 22

TLBoard::TLBoard(bool graphical, GiftPool *gp,
				 QWidget *parent, const char *name)
: Board(TL_WIDTH, TL_HEIGHT, graphical, gp, parent, name)
{
	filled.resize(matrix().height());
}

void TLBoard::init(int seed)
{
	Board::init(seed);
	filled.fill(0);
	addRemoved = 0;
}

bool TLBoard::needRemoving()
{
	for(uint j=0; j<firstClearLine(); j++)
		if ( filled[j]==matrix().width() ) return TRUE;
	return FALSE;
}

void TLBoard::gluePiece()
{
	for(uint k=0; k<currentPiece()->nbBlocks(); k++) {
		int j = currentPiece()->line(k, currentLine());
		filled[j]++;
	}
	Board::gluePiece();
}

void TLBoard::clearLine(uint line)
{
	for (uint i=0; i<matrix().width(); i++) setAndDrawBlock(i, line, 0);
	filled[line] = 0;
}

void TLBoard::copyLine(uint destLine, uint srcLine)
{
	for (uint i=0; i<matrix().width(); i++)
		setAndDrawBlock(i, destLine, matrix()(i, srcLine));
	filled[destLine] = filled[srcLine];
}

void TLBoard::remove()
{
	// when called there *are* full lines

	uint nbFullLines = 0, j = 0;
	for (uint k=0; k<firstClearLine(); k++) {
		while ( filled[k]==matrix().width() ) {
			nbFullLines++;
			k++;
		}
		if ( k==matrix().height() ) break;
		// paranoia [it seems quite impossible for the top line to be full]
		if ( j!=k ) copyLine(j, k);
		j++;
	}
	updateRemoved(nbRemoved() + nbFullLines);
	addRemoved += nbFullLines;
	
	// Assign score according to level and nb of full lines (gameboy style)
	switch (nbFullLines) {
	 case 0: break;
	 case 1: updateScore( score() + 40   * level() ); break;
	 case 2: updateScore( score() + 100  * level() ); break;
	 case 3: updateScore( score() + 300  * level() ); break;
	 case 4: updateScore( score() + 1200 * level() ); break;
	}

	// If we make a multiplum of ten lines, increase level
	if ( (nbRemoved()/10)!=((nbRemoved()-nbFullLines)/10) )
		updateLevel(level()+1);

	// Clear the top
	for (uint j=firstClearLine(); j>firstClearLine() - nbFullLines; j--)
		clearLine(j-1);

	addToClearLines(nbFullLines);
}

/*****************************************************************************/
// Multiplayers methods
uint TLBoard::gift()
{
	uint g = (addRemoved>1 ? addRemoved-1 : 0);
	addRemoved = 0;
	return g;
}

void TLBoard::_putGift(uint nb)
{
	if ( nbClearLines()==0 ) {
		gameOver();
		return;
	}
	
	// we must be careful before lifting the lines : a moving piece can
	// be low enough to be dropped : if possible we lift it of one line
	// else gameover ...
	// we redraw the cleared line carefully (ie only changed colors)
	// so a low piece isn't over-written
	if ( !canPosition(currentPos(), currentLine()-1, currentPiece())
	 	 && moveTo(0, 1)==0 ) {
		gameOver();
		return;
	}
	
	// lift all the cases of one line 
    // (nbClearLine!=0 --> firstClearLine!=height
	for (uint j=firstClearLine(); j>0; j--) copyLine(j, j-1);

	// empty the first line
	clearLine(0);
	
	// fill the emptied low line with garbage :)
	uint _nb = nb;
	uint i;
	do {
		i = randomInt(0, matrix().width()-1, _seed);
		if ( matrix()(i, 0)!=0 ) continue;
		setAndDrawBlock(i, 0, 8);
		_nb--;
	} while ( _nb!=0 );
	filled[0] = nb;
	addToClearLines(-1);
}

void TLBoard::putGift(uint nb)
{
	for (uint k=0; k<nb; k++)
		_putGift(matrix().width() - k - 1);
}

/*****************************************************************************/
// AI
Board *TLBoard::newThinkingBoard()
{
	return new TLBoard(FALSE, 0, this);
}

bool TLBoard::think(const Board &b, uint &rotations, int &dec)
{
	// init board
	partialCopy(b);
	const TLBoard &tlb = (const TLBoard &)b;
	for (uint j=0; j<matrix().height(); j++) filled[j] = tlb.filled[j];
	oneLineDown(); oneLineDown(); // so we can rotate in all cases

	// exit condition
	if ( _cPos==(int)_nbPos ) {
		if ( _cRotation==_nbRotations ) {
			rotations = _betterRotation;
			dec       = _betterDec;
			return TRUE;
		}
		_cPos = 0;
		_cRotation++;
	}

	// rotate
	for (uint i=0; i<_cRotation; i++) rotateLeft();

	// move
	int _dec = -currentPos() - currentPiece()->minX() + _cPos;
	moveTo(_dec, 0);

	// drop
	dropDown();

	// compute points
	int rpoints = thinkingPoints();
	if ( (_cPos==0 && _cRotation==0) || rpoints>_betterPoints) {
		// first calculation or better
		_betterPoints   = rpoints;
		_betterRotation = _cRotation;
		_betterDec      = _dec;
	}

	_cPos++;
	return FALSE;
}

int TLBoard::thinkingPoints()
{
	// nb full lines
	uint nbFullLines = 0;
	for (uint j=0; j<matrix().height(); j++) 
		if ( filled[j]==matrix().width() ) nbFullLines++;

	// here we remove the full lines to calculate accurately the holes
	if ( needRemoving() ) remove();

	// nb filled lines
	uint nbFilledLines = matrix().height() - nbClearLines();

	// nb holes && bv empty spaces in non empty lines
	uint nbHoles = 0, nbSpaces = 0, tmp;
	QArray<bool> nonEmptyColumn(matrix().width());
	bool nonEmptyLine;
	for (uint i=0; i<matrix().width(); i++) nonEmptyColumn[i] = FALSE;
	for (uint j=matrix().height(); j>0; j--) {
		nonEmptyLine = FALSE;
		tmp = 0;
		for (uint i=0; i<matrix().width(); i++) {
			if ( matrix()(i, j-1)!=0 ) {
				if ( !nonEmptyColumn[i] ) nonEmptyColumn[i] = TRUE;
				nonEmptyLine = TRUE;
			} else {
				tmp ++;
				if (nonEmptyColumn[i]) nbHoles++;
			}
		}
		if (nonEmptyLine) nbSpaces += tmp;
	}

	// compute score (#### need tuning ...)
	int points = 0;
	switch (nbFullLines) {
	  case 1: points += 100;  break;
	  case 2: points += 200;  break;
	  case 3: points += 500;  break;
	  case 4: points += 1000; break;	
	}
	points -= nbHoles * 50;
	points -= nbSpaces * 2;

	if ( nbFilledLines<7 )       points -= nbFilledLines;
	else if ( nbFilledLines<13 ) points -= nbFilledLines * 2;
	else if ( nbFilledLines<18 ) points -= nbFilledLines * 3;
	else                         points -= nbFilledLines * 4;
	
	return points;
}

/*****************************************************************************/
// Graphic methods
void TLBoard::drawBlock(QPainter *p, const QRect &r,
						const QColor &col, bool remove) const
{
	if (remove) p->fillRect(r, col);
	else {
		p->setPen( col.light() );
		p->moveTo(r.bottomLeft());
		p->lineTo(r.topLeft());
		p->lineTo(r.topRight());
		
		p->setPen( col.dark() );
		p->moveTo(r.topRight() + QPoint(0,1));
		p->lineTo(r.bottomRight());
		p->lineTo(r.bottomLeft() + QPoint(1,0));
		
		QRect rint(r.topLeft() + QPoint(1,1), r.size() - QSize(2, 2));
		p->fillRect(rint, col);
	}
}

void TLBoard::lightRemoved(bool)
{
	_p.setPen(white);
	_p.setRasterOp(XorROP);
	
	for(uint j=0; j<firstClearLine(); j++) {
		if ( filled[j]!=matrix().width() ) continue;
		for (uint i=0; i<matrix().width(); i++) _p.drawRect(blockRect(i, j));
	}

	_p.setRasterOp(CopyROP);
}
